/* $Id$ */

/*========================================================================
 *  Copyright (c) Michael J. Hammel 1995,1996,1997,1998,1999.
 *========================================================================
 *              FILE NAME: pilotaddr.c
 *            DESCRIPTION: Parser routines for pilot address files
 *      DEFINED CONSTANTS: 
 *       TYPE DEFINITIONS: 
 *      MACRO DEFINITIONS: 
 *       GLOBAL VARIABLES: 
 *       PUBLIC FUNCTIONS: XPPilotParseAddr
 *      PRIVATE FUNCTIONS: ProcessFields SortAddresses SortFields
 *                       : FreeAddressList FreeCategoryList
 *                       : GenerateCategoryList
 *  SOFTWARE DEPENDENCIES: 
 *  HARDWARE DEPENDENCIES: 
 *                  NOTES: 
 *
 * SPECIAL CONSIDERATIONS:
 * Set your tabstops to 3 to make the code more readable.
 *========================================================================
 *
 * MODIFICATION HISTORY:
 * $Log$
 *
 *========================================================================*/
#define XP_CPADDR_C

/* === System Headers === */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>

#ifdef DO_GTK
#include <gtk/gtk.h> 
#else
#include "motif.h"
#endif

/* === Project Headers === */
#include "XNotesPlus.h"
#include "debug.h"

/* === external routines === */

/* === Public routine prototypes === */

/* === Private routine prototypes === */
static void ProcessFields();
static void FreeAddressList();

#ifndef OLD_PILOT_ADDR
static void FreeCategoryList();
static void GenerateCategoryList();
#endif

static void PrintFields();
static void SortAddresses();
static void SortCategoryList();
#ifdef DO_MOTIF
static Boolean SortFields();
#else
static gboolean SortFields();
#endif

/* === Global Variables === */
PilotAddr_T					*Pilot_Address=NULL;
PilotAddrCategory_T		*Pilot_Addr_Categories=NULL;

#ifdef DO_MOTIF
	static Boolean			category_field;
#else
	static gboolean		category_field;
#endif


/*========================================================================
 *	Name:			XPPilotParseAddr
 *	Prototype:	void XPPilotParseAddr(char *filename)
 *					
 *	Description:
 *	Parses the CSV output from pilot-addresses
 *
 *	Input Arguments:
 *		char	*filename		name of file with addresses in it.
 *	Output Arguments:
 *	Return Values:
 *		XP_E_FILE_EXIST		file to parse doesn't exist or is not a text file
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 * I couldn't find any details on the CSV format, so I'm just using the
 * output from pilot-addresses to determine how to parse the input.
 *		
 *========================================================================*/
int 
XPPilotParseAddr(
	char	*filename
)
{
#ifdef DEBUG
	char				*fname="XPPiloParseAddr()";
#endif

#ifdef DO_MOTIF
	Boolean			in_field, escape, skip_home;
#else
	gboolean			in_field, escape, skip_home;
#endif

	struct stat		stat_buf;
	char				buf[4096];
	char				inbuf[5110];
	FILE				*fd;
	char				c;
	int				field_count;
	int				length;
	char				*ptr, *bufptr;
	char				*semicolon, *comma;
	char				*newbuf;

	DBGEnter();
	DBGPrintf(DBG_INFO,("pilot db to look for: %s\n", filename));

	/*
	 * Verify file exists.  Return if it doesn't.
	 */
	if ( stat(filename, &stat_buf) != 0 )
		return XP_E_FILE_EXIST;

	if ( ! ( stat_buf.st_mode & S_IFREG ) )
		return XP_E_FILE_EXIST;

	/*
	 * Free up any old link lists.
	 */
	FreeAddressList();
#ifndef OLD_PILOT_ADDR
	FreeCategoryList();
#endif

	/*
	 * File is ok.  Open it.
	 */
	fd = fopen(filename, "r");

	/*
	 * Here is the ugly part.  Read in the file, one character at a time.
	 * Bleck.  We have to maintain state so we know:
	 * 1. are we in a record or moving to a new record?
	 * 2. are we in a field or at a field delimiter?
	 * 3. does the entry include a category, if not it must be under "Unfiled"
	 */

	in_field = False;
	escape = False;
	skip_home = False;
	category_field = True;
	field_count = 0;
	ptr=buf;

	while ( (fgets(inbuf, 4096, fd)) != NULL ) 
	{
		/*
		 * "Unfiled" entries don't have a category.  Find them, and prefix
		 * them with proper category field data.
		 */
		semicolon = strchr(inbuf, ';');
		comma = strchr(inbuf, ',');
		if ( ( semicolon == NULL ) || ( (int)comma < (int)semicolon ) )
		{
			newbuf = XPSafeAlloc(strlen(inbuf)+11);
			strcpy(newbuf, "\"Unfiled\";");
			strcat(newbuf, inbuf);
			strcpy(inbuf, newbuf);
			free(newbuf);
		}

		length = strlen(inbuf);
		for ( bufptr=inbuf; (int)bufptr < (int)(inbuf+length); bufptr++ )
		{
			c = *bufptr;

			/*
			 * If we're in a field, then we either copy the character to a 
			 * buffer, or if we've found the end of the field we process the
			 * buffer and prep for a new field.
			 */
			if ( in_field )
			{
				switch (c)
				{
	
					/*
					 * Skip escape marker - but flag that the next character is
					 * supposed to be part of the field and not a delimiter.
					 */
					case '\\':	
						escape=True;
						break;
	
					/*
					 * Skip escaped characters, like newline, tab, etc.
					 */
#ifndef OLD_PILOT_ADDR
					case 'n':	
						if ( escape )
						{
							bcopy((char *)"\n", ptr++, 1);
							escape=False;
						}
						else
							bcopy((char *)&c, ptr++, 1);
						break;
	
					case 't':	
						if ( escape )
						{
							bcopy((char *)"\t", ptr++, 1);
							escape=False;
						}
						else
							bcopy((char *)&c, ptr++, 1);
						break;
	
					case 'v':	
						if ( escape )
						{
							bcopy((char *)"\v", ptr++, 1);
							escape=False;
						}
						else
							bcopy((char *)&c, ptr++, 1);
						break;
	
					case 'r':	
						if ( escape )
						{
							bcopy((char *)"\r", ptr++, 1);
							escape=False;
						}
						else
							bcopy((char *)&c, ptr++, 1);
						break;
	
					case 'f':	
						if ( escape )
						{
							bcopy((char *)"\f", ptr++, 1);
							escape=False;
						}
						else
							bcopy((char *)&c, ptr++, 1);
						break;
	
					case 'b':	
						if ( escape )
						{
							bcopy((char *)"\b", ptr++, 1);
							escape=False;
						}
						else
							bcopy((char *)&c, ptr++, 1);
						break;
#endif
	
					/*
					 * If the character is a double quote and it has not been
					 * escaped then we're at the end of a field.  If it has
					 * been escaped, copy to the field buffer.
					 */
					case '\"':	
						if ( escape )
						{
							bcopy((char *)&c, ptr++, 1);
							escape=False;
						}
						else
						{
							bcopy("\0", ptr++, 1);
							in_field = False; 
	
							/*
							 * Leave the Home, Fax, Other, etc fields alone.
							 */
							if ((	( strcmp(buf, "Work"   ) == 0 ) ||
									( strcmp(buf, "Home"   ) == 0 ) ||
									( strcmp(buf, "Fax"    ) == 0 ) ||
									( strcmp(buf, "Other"  ) == 0 ) ||
									( strcmp(buf, "E-mail" ) == 0 ) ||
									( strcmp(buf, "Main"   ) == 0 ) ||
									( strcmp(buf, "Pager"  ) == 0 ) ||
									( strcmp(buf, "Mobile" ) == 0 ) ) && 
									!category_field )
							{
								skip_home = True;
	
								/*
								 * Skip the E-mail and Home tags - they seem broke in 
								 * the Pilot-Link 0.88.1 release.  Or I just don't
								 * understand why they get stuck where they do.
								 */
								if (	( strcmp(buf, "E-mail" ) != 0 ) && 
										( strcmp(buf, "Home"   ) != 0 ) )
								{
									strcat(buf, ": ");
									ptr++;
								}
								else
									ptr=buf;
							}
							else
							{
								skip_home = False;
#ifdef OLD_PILOT_ADDR
								if (field_count > 20 )
#else
								if (field_count > 21 )
#endif
									field_count = 1;
								ProcessFields(field_count, buf);
								ptr=buf;
							}
						}
						break;
	
	
					/*
					 * Everything else is copied to the field buffer.
					 */
					default: 
						bcopy((char *)&c, ptr++, 1);
						break;
				}
			}
			else
			{
				/*
				 * Not in a field yet - look for start of field marker (a 
				 * double quote).
				 */
				switch (c)
				{
	
					/*
					 * The double quote marks the start of a new field.
					 */
					case '\"':	
						in_field = True; 
						if ( ! skip_home )
							field_count++;
						break;
	
					/*
					 * I think that Private (or category) can be just plain
					 * 0 (no double quotes).  Here is where I'd handle that.
					 */
	
	
					/*
					 * Toss out anything else (probably just commas).
					 */
					default:
						break;
				}
			}
		}
	}

	/*
	 * Sort the link list.
	 */
	SortAddresses();

	/*
	 * Generate the link list of categories.
	 */
#ifndef OLD_PILOT_ADDR
	GenerateCategoryList();
	SortCategoryList();
#endif

#ifdef DEBUG
	PrintFields();
#endif
	return ( XP_E_SUCCESS );
}


/*========================================================================
 *	Name:			ProcessFields
 *	Prototype:	void ProcessFields(int count, char *field)
 *					
 *	Description:
 *	Process each field as it is passed in by adding it to the appropriate
 * field of an allocated record.  Allocate records as needed.
 *
 *	Input Arguments:
 *		int	count				field identifier
 *		char	*field		field data
 *========================================================================*/
static void
ProcessFields(
	int	count,
	char	*field
)
{
#ifdef DEBUG
	char				*fname="ProcessFields()";
#endif


	PilotAddr_T					*pentry;
	static PilotAddr_T		*current_link=NULL;

/*
	DBGEnter();
*/

	switch(count)
	{
		case 1: 
			/*
			 * New record.  Allocate a new data structure for it, then
			 * initialize the record.
			 */
			DBGPrintf(DBG_INFO,("Start new record: %s\n", field));
			pentry = (PilotAddr_T *)XPSafeAlloc(sizeof(PilotAddr_T));
			pentry->category = NULL;
			pentry->last = NULL;
			pentry->first = NULL;
			pentry->title = NULL;
			pentry->company = NULL;
			pentry->work = NULL;
			pentry->home = NULL;
			pentry->fax = NULL;
			pentry->other = NULL;
			pentry->email = NULL;
			pentry->address = NULL;
			pentry->city = NULL;
			pentry->state = NULL;
			pentry->zip = NULL;
			pentry->country = NULL;
			pentry->custom1 = NULL;
			pentry->custom2 = NULL;
			pentry->custom3 = NULL;
			pentry->custom4 = NULL;
			pentry->note = NULL;
			pentry->private = NULL;
				
			if ( Pilot_Address == NULL )
				Pilot_Address = pentry;
			else
				current_link->next = pentry;

			current_link = pentry;
			current_link->next = NULL;

#ifdef OLD_PILOT_ADDR
			/* Fill in the last name field */
			current_link->last = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->last, field);
			DBGPrintf(DBG_INFO,("last: %s\n", field));
#else
			/* Fill in the category field */
			current_link->category = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->category, field);
			DBGPrintf(DBG_INFO,("category: %s\n", field));
			category_field = False;
#endif
			break;

		/*
		 * Fill in the appropriate field of the current record
		 * with the field passed to us.
		 */
#ifdef OLD_PILOT_ADDR
		case 2: 
			DBGPrintf(DBG_INFO,("First name: %s\n", field));
			current_link->first = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->first, field);
			break;

		case 3: 
			current_link->title = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->title, field);
			break;

		case 4: 
			current_link->company = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->company, field);
			break;

		case 5:
			current_link->work = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->work, field);
			break;

		case 6: 
			current_link->home = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->home, field);
			break;

		case 7: 
			current_link->fax = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->fax, field);
			break;

		case 8: 
			current_link->other = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->other, field);
			break;

		case 9: 
			current_link->email = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->email, field);
			break;

		case 10: 
			current_link->address = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->address, field);
			break;

		case 11: 
			current_link->city = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->city, field);
			break;

		case 12: 
			current_link->state = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->state, field);
			break;

		case 13: 
			current_link->zip = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->zip, field);
			break;

		case 14: 
			current_link->country = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->country, field);
			break;

		case 15: 
			current_link->custom1 = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->custom1, field);
			break;

		case 16: 
			current_link->custom2 = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->custom2, field);
			break;

		case 17: 
			current_link->custom3 = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->custom3, field);
			break;

		case 18: 
			current_link->custom4 = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->custom4, field);
			break;

		case 19: 
			current_link->note = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->note, field);
			break;

		case 20: 
			current_link->private = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->private, field);
			break;
#else
		case 2: 
			DBGPrintf(DBG_TRACE2,("last: %s\n", field));
			current_link->last = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->last, field);
			break;

		case 3: 
			DBGPrintf(DBG_TRACE2,("first: %s\n", field));
			current_link->first = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->first, field);
			break;

		case 4: 
			DBGPrintf(DBG_TRACE2,("title: %s\n", field));
			current_link->title = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->title, field);
			break;

		case 5: 
			DBGPrintf(DBG_TRACE2,("company: %s\n", field));
			current_link->company = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->company, field);
			break;

		case 6:
			DBGPrintf(DBG_TRACE2,("work: %s\n", field));
			current_link->work = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->work, field);
			break;

		case 7: 
			DBGPrintf(DBG_TRACE2,("home: %s\n", field));
			current_link->home = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->home, field);
			break;

		case 8: 
			DBGPrintf(DBG_TRACE2,("fax: %s\n", field));
			current_link->fax = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->fax, field);
			break;

		case 9: 
			DBGPrintf(DBG_TRACE2,("other: %s\n", field));
			current_link->other = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->other, field);
			break;

		case 10: 
			DBGPrintf(DBG_TRACE2,("email: %s\n", field));
			current_link->email = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->email, field);
			break;

		case 11: 
			DBGPrintf(DBG_TRACE2,("address: %s\n", field));
			current_link->address = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->address, field);
			break;

		case 12: 
			DBGPrintf(DBG_TRACE2,("city: %s\n", field));
			current_link->city = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->city, field);
			break;

		case 13: 
			DBGPrintf(DBG_TRACE2,("state: %s\n", field));
			current_link->state = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->state, field);
			break;

		case 14: 
			DBGPrintf(DBG_TRACE2,("zip: %s\n", field));
			current_link->zip = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->zip, field);
			break;

		case 15: 
			DBGPrintf(DBG_TRACE2,("country: %s\n", field));
			current_link->country = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->country, field);
			break;

		case 16: 
			DBGPrintf(DBG_TRACE2,("custom1: %s\n", field));
			current_link->custom1 = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->custom1, field);
			break;

		case 17: 
			DBGPrintf(DBG_TRACE2,("custom2: %s\n", field));
			current_link->custom2 = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->custom2, field);
			break;

		case 18: 
			DBGPrintf(DBG_TRACE2,("custom3: %s\n", field));
			current_link->custom3 = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->custom3, field);
			break;

		case 19: 
			DBGPrintf(DBG_TRACE2,("custom4: %s\n", field));
			current_link->custom4 = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->custom4, field);
			break;

		case 20: 
			DBGPrintf(DBG_TRACE2,("custom5: %s\n", field));
			current_link->note = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->note, field);
			break;

		case 21: 
			DBGPrintf(DBG_TRACE2,("custom6: %s\n", field));
			current_link->private = XPSafeAlloc(strlen(field) + 1);
			strcpy(current_link->private, field);
			category_field = True;
			break;
#endif

	}

}

static void
PrintFields()
{
#ifdef DEBUG
	char		*fname="PrintFields()";
#endif

	PilotAddr_T		*current_link;
	
	DBGEnter();

	current_link = Pilot_Address;
	while (current_link != NULL )
	{
		DBGPrintf(DBG_TRACE3, ("New record: \n"));
		DBGWrite(DBG_TRACE3,
			("\tlast, first: %s, %s\n", current_link->last, current_link->first));
#ifndef OLD_PILOT_ADDR
		DBGWrite(DBG_TRACE3,
			("\tcategory   : %s\n", current_link->category));
#endif
		DBGWrite(DBG_TRACE3,
			("\ttitle      : %s\n", current_link->title));
		DBGWrite(DBG_TRACE3,
			("\tcompany    : %s\n", current_link->company));
		DBGWrite(DBG_TRACE3,
			("\twork       : %s\n", current_link->work));
		DBGWrite(DBG_TRACE3,
			("\thome       : %s\n", current_link->home));
		DBGWrite(DBG_TRACE3,
			("\tfax        : %s\n", current_link->fax));
		DBGWrite(DBG_TRACE3,
			("\tother      : %s\n", current_link->other));
		DBGWrite(DBG_TRACE3,
			("\temail      : %s\n", current_link->email));
		DBGWrite(DBG_TRACE3,
			("\taddress    : %s\n", current_link->address));
		DBGWrite(DBG_TRACE3,
			("\tcity       : %s\n", current_link->city));
		DBGWrite(DBG_TRACE3,
			("\tstate      : %s\n", current_link->state));
		DBGWrite(DBG_TRACE3,
			("\tzip        : %s\n", current_link->zip));
		DBGWrite(DBG_TRACE3,
			("\tcountry    : %s\n", current_link->country));
		DBGWrite(DBG_TRACE3,
			("\tcustom1    : %s\n", current_link->custom1));
		DBGWrite(DBG_TRACE3,
			("\tcustom2    : %s\n", current_link->custom2));
		DBGWrite(DBG_TRACE3,
			("\tcustom3    : %s\n", current_link->custom3));
		DBGWrite(DBG_TRACE3,
			("\tcustom4    : %s\n", current_link->custom4));
/*
		DBGWrite(DBG_TRACE3,
			("\tnote       : %s\n", current_link->note));
		DBGWrite(DBG_TRACE3,
			("\t     length: %d\n", strlen(current_link->note)));
		DBGWrite(DBG_TRACE3,
			("\tprivate    : %s\n", current_link->private));
			pentry->category = NULL;
*/

		current_link = current_link->next;
	}
}



/*========================================================================
 *	Name:			SortAddresses
 *	Prototype:	void SortAddresses(int primary, int secondary)
 *					
 *	Description:
 * Sort the address list based on which field is of interest.
 *
 *	Input Arguments:
 *		int	primary				primary field to sort on
 *		int	secondary			secondary field to sort on
 *
 * Notes:
 * Uses a simple bubble sort, which should be sufficient for most address
 * lists.  We first sort on the last name, then run the list again to sort
 * by first name.  This only really works if the data in the Pilot is
 * correctly entered (ie the Last Name entry in the pilot contains *only*
 * the Last Name and *not* the last name and first name seperated by a
 * comma).
 *========================================================================*/
static void
SortAddresses(
	int	primary,
	int	secondary
)
{
#ifdef DEBUG
	char				*fname="SortAddresses()";
#endif

	PilotAddr_T		*current_link, *prev_link;

#ifdef DO_MOTIF
	Boolean			repeat, rc1, rc2;
#else
	gboolean			repeat, rc1, rc2;
#endif

	DBGEnter();

repeat_primary:
	repeat = False;
	current_link = Pilot_Address;
	prev_link = NULL;
	while (current_link != NULL)
	{
		rc1 = False;
		if ( current_link->next != NULL )
			rc1 = SortFields(SORT_BY_LAST, current_link, prev_link);
		if ( rc1 )
			repeat = True;
		prev_link = current_link;
		current_link = current_link->next;
	}
	if ( repeat )
		goto repeat_primary;

repeat_secondary:
	repeat = False;
	current_link = Pilot_Address;
	prev_link = NULL;
	while (current_link != NULL)
	{
		rc2 = False;
		if ( current_link->next != NULL )
			if ( strcmp(current_link->last, current_link->next->last) == 0 )
				rc2 = SortFields(SORT_BY_FIRST, current_link, prev_link);
		if ( rc2 )
			repeat = True;
		prev_link = current_link;
		current_link = current_link->next;
	}
	if ( repeat )
		goto repeat_secondary;
}



/*========================================================================
 *	Name:			SortFields
 *	Prototype:	static Boolean SortFields(int field)
 *					
 *	Description:
 * Sort two entries.
 *
 *	Input Arguments:
 *		int				field					field to sort on
 *		PilotAddr_T		*current_link		current link to sort
 *
 * Returns
 *		True				switch occured
 *		False				switch did not occur
 *
 * Notes:
 *	Assumes current_link and current_link->next exist (are non-NULL)
 *========================================================================*/
#ifdef DO_MOTIF
static Boolean
#else
static gboolean
#endif
SortFields(
	int				field,
	PilotAddr_T		*current_link,
	PilotAddr_T		*prev_link
)
{
#ifdef DEBUG
	char				*fname="SortFields()";
#endif

	PilotAddr_T		*next_link;

#ifdef DO_MOTIF
	Boolean			swap;
#else
	gboolean			swap;
#endif

	DBGEnter();

	swap = False;
	next_link = current_link->next;
	if ( next_link == NULL )
		return(False);

	switch ( field )
	{
		case SORT_BY_LAST:
			if ( (current_link->last != NULL) && (next_link->last == NULL) )
					swap = True;
			if ( (current_link->last != NULL) && (next_link->last != NULL) )
				if ( strcmp(current_link->last, next_link->last) > 0 )
					swap = True;
			break;
		case SORT_BY_FIRST:
			if ( (current_link->first != NULL) && (next_link->first == NULL) )
					swap = True;
			if ( (current_link->first != NULL) && (next_link->first != NULL) )
				if ( strcmp(current_link->first, next_link->first) > 0 )
					swap = True;
			break;
		case SORT_BY_COMPANY:
			if ( (current_link->company != NULL) && (next_link->company == NULL) )
					swap = True;
			if ( (current_link->company != NULL) && (next_link->company != NULL) )
				if ( strcmp(current_link->company, next_link->company) > 0 )
					swap = True;
			break;
		case SORT_BY_CATEGORY:
			if ( (current_link->category != NULL) && (next_link->category == NULL) )
					swap = True;
			if ( (current_link->category != NULL) && (next_link->category != NULL) )
				if ( strcmp(current_link->category, next_link->category) > 0 )
					swap = True;
			break;
		case SORT_BY_CITY:
			if ( (current_link->city != NULL) && (next_link->city == NULL) )
					swap = True;
			if ( (current_link->city != NULL) && (next_link->city != NULL) )
				if ( strcmp(current_link->city, next_link->city) > 0 )
					swap = True;
			break;
	}
	if ( swap )
	{
		if ( prev_link != NULL )
			prev_link->next = next_link;
		else
			Pilot_Address = next_link;
		current_link->next = next_link->next;
		next_link->next = current_link;
		return(True);
	}
	return(False);
}


/*========================================================================
 *	Name:			FreeAddressList
 *	Prototype:	static void FreeAddressList()
 *					
 *	Description:
 * Free up the link list of addresses.
 *
 * Notes:
 *========================================================================*/
static void
FreeAddressList()
{
	PilotAddr_T			*current_link, *next_link;

	/*
	 * Free up any old links in the list.
	 */
	current_link = Pilot_Address;
	while (current_link != NULL )
	{
		/* Save the next link */
		next_link = current_link->next;

		/* Free up any allocated storage within the link */
		if ( current_link->last != NULL )
			free(current_link->last);
		if ( current_link->first != NULL )
			free(current_link->first);
		if ( current_link->title != NULL )
			free(current_link->title);
		if ( current_link->company != NULL )
			free(current_link->company);
		if ( current_link->work != NULL )
			free(current_link->work);
		if ( current_link->home != NULL )
			free(current_link->home);
		if ( current_link->fax != NULL )
			free(current_link->fax);
		if ( current_link->other != NULL )
			free(current_link->other);
		if ( current_link->email != NULL )
			free(current_link->email);
		if ( current_link->address != NULL )
			free(current_link->address);
		if ( current_link->city != NULL )
			free(current_link->city);
		if ( current_link->state != NULL )
			free(current_link->state);
		if ( current_link->zip != NULL )
			free(current_link->zip);
		if ( current_link->country != NULL )
			free(current_link->country);
		if ( current_link->custom1 != NULL )
			free(current_link->custom1);
		if ( current_link->custom2 != NULL )
			free(current_link->custom2);
		if ( current_link->custom3 != NULL )
			free(current_link->custom3);
		if ( current_link->custom4 != NULL )
			free(current_link->custom4);
		if ( current_link->note != NULL )
			free(current_link->note);
		if ( current_link->private != NULL )
			free(current_link->private);
		if ( current_link->category != NULL )
			free(current_link->category);

		/* free up the current link */
		free(current_link);
		current_link = next_link;
	}
	Pilot_Address=NULL;
}


/*========================================================================
 *	Name:			FreeCategoryList
 *	Prototype:	static void FreeCategoryList()
 *					
 *	Description:
 * Free up the link list of addresses.
 *
 * Notes:
 *========================================================================*/
#ifndef OLD_PILOT_ADDR
static void
FreeCategoryList()
{
	PilotAddrCategory_T		*category_link, *next_link;

	category_link = Pilot_Addr_Categories;
	while (category_link != NULL )
	{
		if ( category_link->name != NULL )
			free(category_link->name);
		next_link = category_link->next;
		free(category_link);
		category_link = next_link;
	}
	Pilot_Addr_Categories = NULL;
}
#endif


/*========================================================================
 *	Name:			SortCategoryList
 *	Prototype:	static void SortCategoryList()
 *					
 *	Description:
 * Sort the list of categories.
 *
 * Notes:
 *========================================================================*/
#ifndef OLD_PILOT_ADDR
static void
SortCategoryList()
{
#ifdef DO_MOTIF
	Boolean			repeat;
#else
	gboolean			repeat;
#endif
	PilotAddrCategory_T		*category_link, *prev_link, *next_link;

repeat_category:
	repeat = False;
	category_link = Pilot_Addr_Categories;
	prev_link = NULL;
	while (category_link != NULL)
	{
		next_link = category_link->next;
		if ( next_link != NULL )
		{
			if ( strcmp(category_link->name, next_link->name) > 0 )
			{
				if ( prev_link != NULL )
					prev_link->next = next_link;
				else
					Pilot_Addr_Categories = next_link;
				category_link->next = next_link->next;
				next_link->next = category_link;
				repeat = True;
			}
		}
		prev_link = category_link;
		category_link = category_link->next;
	}
	if ( repeat )
		goto repeat_category;

}
#endif


/*========================================================================
 *	Name:			GenerateCategoryList
 *	Prototype:	static void GenerateCategoryList()
 *					
 *	Description:
 * Generate the link list of address categories.
 *
 * Notes:
 *========================================================================*/
#ifndef OLD_PILOT_ADDR
static void
GenerateCategoryList()
{
#ifdef DEBUG
	char							*fname="GenerateCategoryList()";
#endif

	PilotAddr_T					*addr_link;
	PilotAddrCategory_T		*category_link, *last_cat_link;

#ifdef DO_MOTIF
	Boolean						found_match;
#else
	gboolean						found_match;
#endif

	DBGEnter();

	addr_link = Pilot_Address;
	while (addr_link != NULL )
	{
		if ( Pilot_Addr_Categories == NULL )
		{
			/* First entry:  allocate an entry and point the header to it. */
			Pilot_Addr_Categories = 
				(PilotAddrCategory_T *)XPSafeAlloc(sizeof(PilotAddrCategory_T));
			Pilot_Addr_Categories->next = NULL;
			Pilot_Addr_Categories->name = 
				XPSafeAlloc(strlen(addr_link->category)+1);
			strcpy(Pilot_Addr_Categories->name, addr_link->category);
		}
		else
		{
			/* Check to see if category already exists in list */
			category_link = Pilot_Addr_Categories;
			found_match = False;
			while (category_link != NULL )
			{
				if ( strcmp(category_link->name, addr_link->category) == 0 )
				{
					found_match = True;
					break;
				}
				last_cat_link = category_link;
				category_link = category_link->next;
			}

			/*
			 * If no match was found, insert a new link at the end of the
			 * current list.  We'll sort this list later.
			 */
			if (!found_match)
			{
				/*
				 * last_cat_link should point to the last link in the current
				 * list.
				 */
				last_cat_link->next = 
					(PilotAddrCategory_T *)XPSafeAlloc(sizeof(PilotAddrCategory_T));
				category_link = last_cat_link->next;
				category_link->next = NULL;
				category_link->name = XPSafeAlloc(strlen(addr_link->category)+1);
				strcpy(category_link->name, addr_link->category);
			}
		}

		/*
		 * Get the next link in the address list.
		 */
		addr_link = addr_link->next;
	}
}
#endif
